@@ -10,14 +10,25 @@ from queue import LifoQueue
 from typing import Any, Dict, Optional, Type, Union
 from uuid import UUID, uuid4
 
+import backoff
 import httpx
 import jwt
 import structlog
 import websockets
+from httpx import ReadTimeout
+from nbclient.util import run_sync
 from pydantic import BaseModel, BaseSettings, ValidationError
 
+from origami.types.rtu import BulkCellStateMessage
+
 from .types.deltas import FileDeltaAction, FileDeltaType, NBCellProperties, V2CellContentsProperties
-from .types.files import NotebookFile
+from .types.files import FileVersion, NotebookFile
+from .types.jobs import (
+    CreateParameterizedNotebookRequest,
+    CustomerJobInstanceReference,
+    CustomerJobInstanceReferenceInput,
+    JobInstanceAttempt,
+)
 from .types.kernels import SessionRequestDetails
 from .types.rtu import (
     RTU_ERROR_HARD_MESSAGE_TYPES,
@@ -26,7 +37,6 @@ from .types.rtu import (
     AuthenticationRequest,
     AuthenticationRequestData,
     CallbackTracker,
-    CellStateMessageReply,
     FileDeltaReply,
     FileSubscribeReplySchema,
     GenericRTUMessage,
@@ -122,20 +132,24 @@ class NoteableClient(httpx.AsyncClient):
         if not config:
             settings = ClientSettings()
             if not os.path.exists(settings.auth0_config_path):
-                logger.error(
+                logger.warning(
                     f"No config object passed in and no config file found at {settings.auth0_config_path}"
                     ", using default empty config"
                 )
                 config = ClientConfig()
             else:
                 config = ClientConfig.parse_file(settings.auth0_config_path)
+
         self.config = config
+        self.config.domain = os.getenv(
+            "NOTEABLE_URI", os.getenv("NOTEABLE_DOMAIN", self.config.domain)
+        )
         self.file_session_cache = {}
 
         self.user = None
-        self.token = api_token or self.get_token()
+        self.token = api_token or os.getenv("NOTEABLE_TOKEN") or self.get_token()
         if isinstance(self.token, str):
-            self.token = Token(access_token=api_token)
+            self.token = Token(access_token=self.token)
         self.rtu_socket = None
         self.process_task_loop = None
 
@@ -189,12 +203,27 @@ class NoteableClient(httpx.AsyncClient):
         token_data = jwt.decode(token, options={"verify_signature": False})
         return Token(access_token=token, **token_data)
 
-    async def get_notebook(self, file_id) -> NotebookFile:
+    @_default_timeout_arg
+    @backoff.on_exception(backoff.expo, ReadTimeout, max_time=10)
+    async def get_notebook(self, file_id, timeout=None) -> NotebookFile:
         """Fetches a notebook file via the Noteable REST API as a NotebookFile model (see files.py)"""
-        resp = await self.get(f"{self.api_server_uri}/files/{file_id}")
+        resp = await self.get(f"{self.api_server_uri}/files/{file_id}", timeout=timeout)
         resp.raise_for_status()
         return NotebookFile.parse_raw(resp.content)
 
+    @_default_timeout_arg
+    @backoff.on_exception(backoff.expo, ReadTimeout, max_time=10)
+    async def get_version_or_none(
+        self, version_id: UUID, timeout: int = None
+    ) -> Optional[FileVersion]:
+        """Fetches a file version via the Noteable REST API as a FileVersion model (see files.py)"""
+        resp = await self.get(f"{self.api_server_uri}/fileversions/{version_id}", timeout=timeout)
+        if resp.status_code == 404:
+            return
+        resp.raise_for_status()
+        return FileVersion.parse_raw(resp.content)
+
+    @backoff.on_exception(backoff.expo, ReadTimeout, max_time=10)
     async def get_kernel_session(
         self, file: Union[UUID, NotebookFile]
     ) -> Optional[KernelStatusUpdate]:
@@ -223,7 +252,7 @@ class NoteableClient(httpx.AsyncClient):
             file, kernel_name=kernel_name, hardware_size=hardware_size
         )
         # Needs the .dict conversion to avoid thinking it's an object with a synchronous byte stream
-        resp = await self.post(f"{self.api_server_uri}/sessions", data=request.json())
+        resp = await self.post(f"{self.api_server_uri}/v1/sessions", content=request.json())
         resp.raise_for_status()
         resp_data = resp.json()
         session = KernelStatusUpdate(session_id=resp_data["id"], kernel=resp_data["kernel"])
@@ -294,6 +323,51 @@ class NoteableClient(httpx.AsyncClient):
         if file_id in self.file_session_cache:
             del self.file_session_cache[file.id]
 
+    @_default_timeout_arg
+    async def create_parameterized_notebook(
+        self,
+        notebook_id: UUID,
+        job_instance_attempt: JobInstanceAttempt = None,
+        timeout: float = None,
+    ):
+        """
+        Creates a parameterized_notebook given a notebook version id or a notebook file id.
+
+        If given a notebook version id, fetch the version details and extract file id to construct the request path.
+        """
+        file_version = await self.get_version_or_none(notebook_id)
+        file_id = notebook_id if file_version is None else file_version.file_id
+        notebook_version_id = None if file_version is None else file_version.id
+        body = CreateParameterizedNotebookRequest(
+            notebook_version_id=notebook_version_id, job_instance_attempt=job_instance_attempt
+        )
+        resp = await self.post(
+            f"{self.api_server_uri}/v1/files/{file_id}/parameterized_notebooks",
+            content=body.json(),
+            timeout=timeout,
+        )
+        resp.raise_for_status()
+        file: NotebookFile = NotebookFile.parse_obj(resp.json())
+        file.content = httpx.get(file.presigned_download_url).content.decode("utf-8")
+        return file
+
+    @_default_timeout_arg
+    async def create_job_instance(
+        self, job_instance_input: CustomerJobInstanceReferenceInput, timeout: float = None
+    ) -> CustomerJobInstanceReference:
+        """Create a job instance in Noteable, reference a job instance in a third-party system.
+
+        A job definition has many job instances, a job instance has many attempts.
+        Each attempt is created through the `create_parameterized_notebook` method.
+        """
+        resp = await self.post(
+            f"{self.api_server_uri}/v1/customer-job-instances",
+            content=job_instance_input.json(exclude_unset=True),
+            timeout=timeout,
+        )
+        resp.raise_for_status()
+        return CustomerJobInstanceReference.parse_obj(resp.json())
+
     @property
     def in_context(self):
         """Indicates if the client is within an async context generation loop or not."""
@@ -337,6 +411,12 @@ class NoteableClient(httpx.AsyncClient):
         finally:
             return await httpx.AsyncClient.__aexit__(self, exc_type, exc, tb)
 
+    def __enter__(self):
+        return run_sync(self.__aenter__)()
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        return run_sync(self.__aexit__)(exc_type, exc_val, exc_tb)
+
     def register_message_callback(
         self,
         callable: RTUEventCallable,
@@ -536,7 +616,7 @@ class NoteableClient(httpx.AsyncClient):
             # before the subscribe_reply (for e.g. update_user_file_subscription_event)
             if resp.event != "subscribe_reply":
                 raise SkipCallback("This callback only processes subscribe_reply")
-            resp = GenericRTUReplySchema[TopicActionReplyData].parse_obj(resp)
+            resp = FileSubscribeReplySchema.parse_obj(resp)
             if resp.data.success:
                 self.subscriptions.add(resp.channel)
             else:
@@ -584,10 +664,8 @@ class NoteableClient(httpx.AsyncClient):
         else:
             file_id = file
             # from_delta_id = from_delta_id
-            from_version_id = file.current_version_id
         channel = self.files_channel(file_id)
         req, tracker = self._gen_subscription_request(channel)
-        tracker.response_schema = FileSubscribeReplySchema
         # TODO: write test for these fields
         req.data = {}
         if from_version_id:
@@ -714,12 +792,13 @@ class NoteableClient(httpx.AsyncClient):
         tracker_future = tracker.next_trigger
         results_tracker_future = None
 
-        async def cell_complete_check(resp: CellStateMessageReply):
-            if resp.data.cell_id != cell_id:
-                raise SkipCallback("Not tracked cell")
-            if not resp.data.state.is_terminal_state:
-                raise SkipCallback("Not terminal state")
-            return resp
+        async def cell_complete_check(resp: BulkCellStateMessage):
+            for cell_state in resp.data.cell_states:
+                if cell_state.cell_id == cell_id and cell_state.state.is_terminal_state:
+                    return cell_state
+                elif cell_state.cell_id == cell_id and not cell_state.state.is_terminal_state:
+                    raise SkipCallback("Not terminal state")
+            raise SkipCallback("Not tracked cell")
 
         if await_results:
             assert (
@@ -729,8 +808,8 @@ class NoteableClient(httpx.AsyncClient):
             results_tracker = self.register_message_callback(
                 cell_complete_check,
                 session.kernel_channel,
-                "cell_state_update_event",
-                response_schema=CellStateMessageReply,
+                "bulk_cell_state_update_event",
+                response_schema=BulkCellStateMessage,
             )
             results_tracker_future = results_tracker.next_trigger
 
